# Type Scaffolding in Estructura

This guide explains Estructura's type scaffolding system, which provides a powerful way to create custom type modules with consistent validation, coercion, and generation capabilities.

## What is Type Scaffolding?

Type scaffolding in Estructura is a framework that allows you to:

1. Define reusable type patterns that implement the `Estructura.Nested.Type` behaviour
2. Generate custom types with consistent interfaces
3. Reduce boilerplate when working with similar data types
4. Apply consistent validation, coercion, and generator logic

Type scaffolds are metaprogramming tools that generate the implementation for custom types, ensuring they integrate properly with Estructura's nested structures.

## Core Concepts

### The Type Behaviour

All types in Estructura, including scaffolded types, implement the `Estructura.Nested.Type` behaviour, which requires:

```elixir
defmodule Estructura.Nested.Type do
  @callback generate(opts :: Keyword.t()) :: StreamData.t()
  @callback coerce(term()) :: {:ok, term()} | {:error, String.t()}
  @callback validate(term()) :: {:ok, term()} | {:error, String.t()}
end
```

- `generate/1` - Creates sample values for property testing
- `coerce/1` - Converts input values to the expected type
- `validate/1` - Ensures values meet type constraints

### The Scaffold Pattern

Type scaffolds follow this pattern:

1. Define common behaviors for a family of types
2. Accept configuration options to customize the generated type
3. Generate all necessary code for consistent implementations
4. Provide a common interface for working with the types

## Using Built-in Scaffolds

### Enum Scaffold

The `Estructura.Nested.Type.Enum` scaffold creates types that can only contain predetermined values (similar to enums in other languages).

#### Basic Usage

```elixir
defmodule Status do
  use Estructura.Nested.Type.Enum, elements: [:pending, :active, :completed]
end

Status.validate(:pending)       #=> {:ok, :pending}
Status.validate(:invalid)       #=> {:error, "Expected :invalid to be one of: [:pending, :active, :completed]"}
Status.generate() |> Enum.take(1) #=> [:active] (random element)
```

#### Custom Coercion

You can customize how values are coerced:

```elixir
defmodule Priority do
  use Estructura.Nested.Type.Enum,
    elements: [:low, :medium, :high],
    coercer: fn
      str when is_binary(str) -> {:ok, String.to_existing_atom(str)}
      atom when is_atom(atom) -> {:ok, atom}
      other -> {:error, "Cannot coerce #{inspect(other)} to priority"}
    end
end

Priority.coerce("medium")  #=> {:ok, :medium}
Priority.coerce(:high)     #=> {:ok, :high}
Priority.coerce(123)       #=> {:error, "Cannot coerce 123 to priority"}
```

#### Custom JSON Encoding

Add custom JSON encoding for your enum type:

```elixir
defmodule Role do
  use Estructura.Nested.Type.Enum,
    elements: [:admin, :user, :guest],
    encoder: fn role, opts -> Jason.Encode.string(Atom.to_string(role), opts) end
end

Jason.encode!(:admin)  #=> "\"admin\""
```

### Tags Scaffold

The `Estructura.Nested.Type.Tags` scaffold creates types for handling lists of predefined tags, ensuring they're valid and unique.

#### Basic Usage

```elixir
defmodule Categories do
  use Estructura.Nested.Type.Tags, elements: [:tech, :art, :science]
end

Categories.validate([:tech, :art])   #=> {:ok, [:tech, :art]}
Categories.validate([:invalid])       #=> {:error, "All tags are expected to be one of [:tech, :art, :science]..."}
Categories.generate() |> Enum.take(1) #=> [[:tech]] (random subset)
```

#### Custom Coercion

Add custom coercion logic for tag lists:

```elixir
defmodule ContentTags do
  use Estructura.Nested.Type.Tags,
    elements: [:blog, :news, :tutorial],
    coercer: fn
      tags when is_list(tags) ->
        {:ok, Enum.map(tags, fn
          tag when is_binary(tag) -> String.to_existing_atom(tag)
          tag when is_atom(tag) -> tag
        end) |> Enum.uniq()}
      other -> {:error, "Cannot coerce #{inspect(other)} to tag list"}
    end
end

ContentTags.coerce(["blog", "news"])  #=> {:ok, [:blog, :news]}
ContentTags.coerce([:blog, :blog])    #=> {:ok, [:blog]} (duplicate removed)
```

## Creating Custom Type Scaffolds

You can create your own type scaffolds for reusable type patterns in your application.

### Step 1: Define the Scaffold Module

```elixir
defmodule MyApp.Type.Range do
  @moduledoc """
  A type scaffold for ranges with minimum and maximum values.
  """
  
  @behaviour Estructura.Nested.Type.Scaffold

  @impl true
  def type_module_ast(name, opts) when is_list(opts) do
    defmodule name do
      @moduledoc false
      @opts opts

      require MyApp.Type.Range.Gen
      MyApp.Type.Range.Gen.type_module_ast(@opts)
    end
  end

  defmacro __using__(opts) do
    quote do
      require MyApp.Type.Range.Gen
      MyApp.Type.Range.Gen.type_module_ast(unquote(opts))
    end
  end

  defmodule Gen do
    @moduledoc false
    
    defmacro type_module_ast(opts) do
      opts = Macro.expand(opts, __CALLER__)
      
      min = Keyword.fetch!(opts, :min)
      max = Keyword.fetch!(opts, :max)
      coercer = Keyword.get(opts, :coercer)
      encoder = Keyword.get(opts, :encoder)
      
      quote generated: true, location: :keep do
        @moduledoc "Range type with values between #{unquote(min)} and #{unquote(max)}"
        @behaviour Estructura.Nested.Type
        
        @impl true
        def generate(opts \\ []) do
          StreamData.integer(unquote(min)..unquote(max))
        end
        
        @impl true
        case unquote(coercer) do
          fun when is_function(fun, 1) ->
            def coerce(term), do: unquote(coercer).(term)
            
          nil ->
            def coerce(term) when is_integer(term) do
              validate(term)
            end
            
            def coerce(term) when is_binary(term) do
              case Integer.parse(term) do
                {num, ""} -> validate(num)
                _ -> {:error, "Cannot coerce #{inspect(term)} to integer"}
              end
            end
            
            def coerce(term), do: {:error, "Cannot coerce #{inspect(term)} to integer"}
        end
        
        @impl true
        def validate(term) when is_integer(term) and term >= unquote(min) and term <= unquote(max) do
          {:ok, term}
        end
        
        def validate(term) when is_integer(term) do
          {:error, "Value #{inspect(term)} outside allowed range #{unquote(min)}..#{unquote(max)}"}
        end
        
        def validate(term) do
          {:error, "Expected integer, got: #{inspect(term)}"}
        end
        
        if Code.ensure_loaded?(Jason.Encoder) and is_function(unquote(encoder), 2) do
          defimpl Jason.Encoder do
            @moduledoc false
            def encode(term, opts), do: unquote(encoder).(term, opts)
          end
        end
      end
    end
  end
end
```

### Step 2: Use Your Custom Scaffold

```elixir
defmodule MyApp.Types.Age do
  use MyApp.Type.Range, min: 0, max: 120
end

defmodule MyApp.Types.Rating do
  use MyApp.Type.Range, 
    min: 1, 
    max: 5,
    coercer: fn
      n when is_float(n) -> {:ok, round(n)}
      n when is_integer(n) -> {:ok, n}
      other -> {:error, "Cannot coerce #{inspect(other)} to rating"}
    end
end

MyApp.Types.Age.validate(30)      #=> {:ok, 30}
MyApp.Types.Age.validate(150)     #=> {:error, "Value 150 outside allowed range 0..120"}
MyApp.Types.Age.coerce("42")      #=> {:ok, 42}

MyApp.Types.Rating.validate(5)    #=> {:ok, 5}
MyApp.Types.Rating.coerce(4.7)    #=> {:ok, 5}
```

## Best Practices for Working with Scaffolded Types

### 1. Organize Types Logically

Group related types together based on their domain:

```elixir
defmodule MyApp.Types do
  defmodule Status do
    use Estructura.Nested.Type.Enum, elements: [:pending, :active, :completed]
  end
  
  defmodule Priority do
    use Estructura.Nested.Type.Enum, elements: [:low, :medium, :high]
  end
  
  defmodule Tags do
    use Estructura.Nested.Type.Tags, elements: [:feature, :bug, :doc, :test]
  end
end
```

### 2. Document Your Types

Add proper documentation to your type modules:

```elixir
defmodule MyApp.Types.Status do
  @moduledoc """
  Represents the current status of an item in the system.
  
  Possible values:
  - `:pending` - Item has been created but processing hasn't started
  - `:active` - Item is currently being processed
  - `:completed` - Item has finished processing
  """
  use Estructura.Nested.Type.Enum, elements: [:pending, :active, :completed]
end
```

### 3. Use Consistent Coercion Strategies

Define consistent coercion patterns across your types:

```elixir
# String to atom coercion helper
defmodule MyApp.Types.Helpers do
  def string_to_atom_coercer(elements) do
    fn
      val when is_atom(val) and val in elements -> {:ok, val}
      val when is_binary(val) ->
        try do
          atom = String.to_existing_atom(val)
          if atom in elements, do: {:ok, atom}, else: {:error, "Invalid value: #{val}"}
        rescue
          ArgumentError -> {:error, "Unknown value: #{val}"}
        end
      other -> {:error, "Cannot coerce #{inspect(other)} to enum"}
    end
  end
end

defmodule MyApp.Types.Status do
  @elements [:pending, :active, :completed]
  use Estructura.Nested.Type.Enum,
    elements: @elements,
    coercer: MyApp.Types.Helpers.string_to_atom_coercer(@elements)
end
```

### 4. Handle Invalid Values Gracefully

Provide clear error messages in your custom coercers:

```elixir
defmodule MyApp.Types.Rating do
  use MyApp.Type.Range, 
    min: 1, 
    max: 5,
    coercer: fn
      n when is_float(n) and n >= 1 and n <= 5 -> 
        {:ok, round(n)}
      n when is_integer(n) and n >= 1 and n <= 5 -> 
        {:ok, n}
      n when is_number(n) -> 
        {:error, "Rating must be between 1 and 5, got: #{n}"}
      other -> 
        {:error, "Rating must be a number, got: #{inspect(other)}"}
    end
end
```

### 5. Test Your Types Thoroughly

Use property-based testing to verify your types:

```elixir
defmodule MyApp.Types.StatusTest do
  use ExUnit.Case, async: true
  use ExUnitProperties
  
  alias MyApp.Types.Status
  
  property "generates valid statuses" do
    check all status <- Status.generate() do
      assert status in [:pending, :active, :completed]
      assert {:ok, ^status} = Status.validate(status)
    end
  end
  
  test "validates valid statuses" do
    for status <- [:pending, :active, :completed] do
      assert {:ok, ^status} = Status.validate(status)
    end
  end
  
  test "coerces strings to statuses" do
    for {string, expected} <- [{"pending", :pending}, {"active", :active}, {"completed", :completed}] do
      assert {:ok, ^expected} = Status.coerce(string)
    end
  end
end
```

### 6. Integrate with Estructura.Nested

Use your scaffolded types with nested structures:

```elixir
defmodule MyApp.Task do
  use Estructura.Nested
  
  alias MyApp.Types
  
  defstruct [
    title: "",
    status: :pending,
    priority: :medium,
    tags: []
  ]
  
  def type(:status), do: Types.Status
  def type(:priority), do: Types.Priority 
  def type(:tags), do: Types.Tags
end

# Create and validate a task
task = %MyApp.Task{
  title: "Implement feature",
  status: :active,
  priority: "high",  # will be coerced to :high
  tags: ["feature", "urgent"]  # will be coerced to [:feature, :urgent]
}

{:ok, valid_task} = MyApp.Task.validate(task)
```

## Conclusion

Type scaffolding in Estructura provides a powerful way to create consistent, reusable types with built-in validation, coercion, and generation capabilities. By using the built-in scaffolds or creating your own, you can significantly reduce boilerplate code and ensure type consistency throughout your application.

For more examples and complex use cases, refer to the test files in the Estructura repository.

