searchData={"items":[{"type":"module","title":"Estructura","doc":"`Estructura` is a set of extensions for Elixir structures,\n  such as `Access` implementation, `Enumerable` and `Collectable`\n  implementations, validations and test data generation via `StreamData`.\n\n`Estructura` simplifies the following\n\n  * `Access` implementation for structs\n  * `Enumerable` implementation for structs (as maps)\n  * `Collectable` implementation for one of struct’s fields (as `MapSet` does)\n  * `StreamData` generation of structs for property-based testing","ref":"Estructura.html"},{"type":"module","title":"Use Options - Estructura","doc":"`use Estructura` accepts four keyword arguments.\n\n  * `access: true | false | :lazy` whether to generate the `Access` implementation, default `true`;\n    when `true` or `:lazy`, it also produces `put/3` and `get/3` methods to be used with `coercion`\n    and `validation`, when `:lazy`, instances of `Estructura.Lazy` are understood as values\n  * `coercion: boolean() | [key()]` whether to generate the bunch of `coerce_×××/1` functions\n    to be overwritten by implementations, default `false`\n  * `validation: boolean() | [key()]` whether to generate the bunch of `validate_×××/1` functions\n    to be overwritten by implementations, default `false`\n  * `calculated: [{key(), formula}] when formula: binary() | Formulae.t() | (t() -> any())` the calculated fields \n  * `enumerable: boolean()` whether to generate the `Enumerable` porotocol implementation, default `false`\n  * `collectable: false | key()` whether to generate the `Collectable` protocol implementation,\n    default `false`; if non-falsey atom is given, it must point to a struct field where `Collectable`\n    would collect. Should be one of `list()`, `map()`, `MapSet.t()`, `bitstribg()`\n  * `generator: %{optional(key()) => Estructura.Config.generator()}` the instructions\n    for the `__generate__/{0,1}` functions that would produce the target structure values suitable\n    for usage in `StreamData` property testing; the generated `__generator__/1` function is overwritable.\n\nPlease note, that setting `coercion` and/or `validation` to truthy values has effect\n  if and only if `access` has been also set to `true`.\n\nTypical example of usage would be:\n\n```elixir\ndefmodule MyStruct do\n  use Estructura,\n    access: true,\n    coercion: [:foo], # requires `c:MyStruct.Coercible.coerce_foo/1` impl\n    validation: true, # requires `c:MyStruct.Validatable.validate_×××/1` impls\n    calculated: [foo: \"length(bar)\"], # requires `:formulae` dependency\n    enumerable: true,\n    collectable: :bar,\n    generator: [\n      foo: {StreamData, :integer},\n      bar: {StreamData, :list_of, [{StreamData, :string, [:alphanumeric]}]},\n      baz: {StreamData, :fixed_map,\n        [[key1: {StreamData, :integer}, key2: {StreamData, :integer}]]}\n    ]\n\n  defstruct foo: 0, bar: [], baz: %{}\n\n  @impl MyStruct.Coercible\n  def coerce_foo(value) when is_integer(value), do: {:ok, value}\n  def coerce_foo(value) when is_float(value), do: {:ok, round(value)}\n  def coerce_foo(value) when is_binary(value) do\n    case Integer.parse(value) do\n      {value, \"\"} -> {:ok, value}\n      _ -> {:error, \"#{value} is not a valid integer value\"}\n    end\n  end\n  def coerce_foo(value),\n    do: {:error, \"Cannot coerce value given for `foo` field (#{inspect(value)})\"}\n\n  @impl MyStruct.Validatable\n  def validate_foo(value) when value >= 0, do: {:ok, value}\n  def validate_foo(_), do: {:error, \":foo must be positive\"}\n\n  @impl MyStruct.Validatable\n  def validate_bar(value), do: {:ok, value}\n\n  @impl MyStruct.Validatable\n  def validate_baz(value), do: {:ok, value}\nend\n```\n\nThe above would allow the following to be done with the structure:\n\n```elixir\ns = %MyStruct{}\n\nput_in s, [:foo], \"42\"\n#⇒ %MyStruct{foo: 42, bar: [], baz: %{}}\n\nfor i <- [1, 2, 3], into: s, do: i\n#⇒ %MyStruct{foo: 0, bar: [1, 2, 3], baz: %{}}\n\nEnum.map(s, &elem(&1, 1))\n#⇒ [0, [], %{}]\n\nMyStruct.__generator__() |> Enum.take(3)\n#⇒ [\n#      %MyStruct{bar: [], baz: %{key1: 0, key2: 0}, foo: -1},\n#      %MyStruct{bar: [\"g\", \"xO\"], baz: %{key1: -1, key2: -2}, foo: 2},\n#      %MyStruct{bar: [\"\", \"\", \"\"], baz: %{key1: -3, key2: 1}, foo: -1}\n#    ]\n```","ref":"Estructura.html#module-use-options"},{"type":"module","title":"Calculated fields - Estructura","doc":"When using `Access`, the calculated fields would be also updated upon the\n  update the fields then depend on.","ref":"Estructura.html#module-calculated-fields"},{"type":"module","title":"Coercion - Estructura","doc":"When `coercion: true | [key()]` is passed as an argument to `use Estructura`,\nthe nested behaviour `Coercible` is generated and the target module claims to implement it.\n\nTo make a coercion work with `MyStruct.put/3` and `put_in/3` provided\nby `Access` implementation, the consumer module should implement `MyStruct.Coercible`\nbehaviour.\n\nFor the consumer convenience, the warnings for not implemented functions will be issued by compiler.","ref":"Estructura.html#module-coercion"},{"type":"module","title":"Validation - Estructura","doc":"When `validation: true | [key()]` is passed as an argument to `use Estructura`,\nthe nested behaviour `Validatable` is generated and the target module claims to implement it.\n\nTo make a validation work with `MyStruct.put/3` and `put_in/3` provided\nby `Access` implementation, the consumer module should implement `MyStruct.Validatable`\nbehaviour.\n\nFor the consumer convenience, the warnings for not implemented functions will be issued by compiler.","ref":"Estructura.html#module-validation"},{"type":"module","title":"Generation - Estructura","doc":"If `generator` keyword argument has been passed, `MyStruct.__generate__/{0,1}` can be\nused to generate instances of this struct for `StreamData` property based tests.\n\n```elixir\nproperty \"generation\" do\n  check all %MyStruct{foo: foo, bar: bar, baz: baz} <- MyStruct.__generator__() do\n    assert match?(%{key1: v1, key2: v2} when is_integer(v1) and is_integer(v2), baz)\n    assert is_integer(foo)\n    assert is_binary(bar)\n  end\nend\n```","ref":"Estructura.html#module-generation"},{"type":"module","title":"Lazy - Estructura","doc":"If `access: :lazy` is passed as an option, the struct content might be instantiated lazily,\nupon first access through `Kernel.×××_in/{2,3}` family.\n\nThis might be explicitly helpful when the real content requires a significant time\nto load and/or store. Consider the full response from the web server, including\nthe gzipped content, which might in turn be a huge text file. Or an attachment to an email.\n\nInstead of unarchiving the content, one might use `Lazy` as\n\n```elixir\ndefmodule Response do\n  @moduledoc false\n  use Estructura, access: :lazy\n\n  def extract(file), do: {:ok, ZipHelper.unzip(file)}\n\n defstruct __lazy_data__: nil,\n   file: Estructura.Lazy.new(&Response.extract/1)\nend\n\nresponse = %Response{__lazy_data__: zipped_content}\n# immediate response\n\nresponse |> get_in([:file])\n# unzip and return\n\n{unzipped, struct_with_cached_value} = response |> pop_in([:file])\n# unzip and return the value, alter the struct with it\n```\n\nSee `Estructura.Lazy` for details and options, see `Estructura.LazyMap` for\nthe implementation of lazy map.","ref":"Estructura.html#module-lazy"},{"type":"function","title":"Estructura.coerce/3","doc":"Instantiates the struct by using `Access` from a map, passing all coercions and validations.","ref":"Estructura.html#coerce/3"},{"type":"function","title":"Estructura.diff/3","doc":"Calculates the difference between two estructures and returns a tuple with\n  the first element containing same values and the second one with diffs.\n\nThis function accepts maps but this options should be used as a last resort\nbecause structs are 4–6 times faster.","ref":"Estructura.html#diff/3"},{"type":"function","title":"Examples - Estructura.diff/3","doc":"```elixir\ndefmodule M do\n  use Estructura, enumerable: true\n  defstruct a: true, b: false\nend\nEstructura.diff(struct(M, []), struct(M, b: true), :diff)\n#⇒{%{a: true}, %{b: [false, true]}}\n\nEstructura.diff(%{a: true, b: false}, %{a: true, b: true}, :overlap)\n#⇒ %{a: true}\n\nEstructura.diff(%{a: true, b: false}, %{a: true, b: true}, :disjoint)\n#⇒ %{b: [false, true]}\n```","ref":"Estructura.html#diff/3-examples"},{"type":"type","title":"Estructura.diff_result/0","doc":"Diff return type","ref":"Estructura.html#t:diff_result/0"},{"type":"module","title":"Estructura.Aston","doc":"The implementation of `Estructura` ready to work with tree AST-like structure","ref":"Estructura.Aston.html"},{"type":"function","title":"Estructura.Aston.__generator__/0","doc":"See `Estructura.Aston.__generator__/1`","ref":"Estructura.Aston.html#__generator__/0"},{"type":"function","title":"Estructura.Aston.__generator__/1","doc":"Returns the generator to be used in `StreamData`-powered property testing, based\n  on the specification given to `use Estructura`, which contained","ref":"Estructura.Aston.html#__generator__/1"},{"type":"function","title":"shape - Estructura.Aston.__generator__/1","doc":"```elixir\n%Estructura.Config{\n  access: true,\n  coercion: true,\n  validation: true,\n  calculated: [],\n  collectable: :content,\n  enumerable: true,\n  generator: [\n    name: {StreamData, :string, [:alphanumeric]},\n    attributes: {StreamData, :map_of,\n     [\n       {StreamData, :atom, [:alphanumeric]},\n       {StreamData, :one_of,\n        [\n          [\n            {StreamData, :integer},\n            {StreamData, :boolean},\n            {StreamData, :string, [:alphanumeric]}\n          ]\n        ]}\n     ]},\n    content: {StreamData, :tree,\n     [{StreamData, :fixed_list, [[]]}, &Estructura.Aston.child_gen/1]}\n  ]\n}\n```\n\n\nThe argument given would be used as a template to generate new values.","ref":"Estructura.Aston.html#__generator__/1-shape"},{"type":"function","title":"Estructura.Aston.access/2","doc":"Returns the key to be used for accessing the nested element(s)","ref":"Estructura.Aston.html#access/2"},{"type":"function","title":"Estructura.Aston.coerce/3","doc":"Coerces the deeply nested map to an instance of nested `Estructura.Aston`","ref":"Estructura.Aston.html#coerce/3"},{"type":"function","title":"Estructura.Aston.get/3","doc":"Gets the value for the given key from the structure","ref":"Estructura.Aston.html#get/3"},{"type":"function","title":"Estructura.Aston.put/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns `{:ok, updated_struct}` or `{:error, reason}` if there is no such key","ref":"Estructura.Aston.html#put/3"},{"type":"function","title":"Estructura.Aston.put!/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns the value or raises if there is no such key","ref":"Estructura.Aston.html#put!/3"},{"type":"function","title":"Estructura.Aston.recalculate_calculated/1","doc":"Recalculates calculated fields for the instance of Estructura.Aston.\n\nNormally one would not need to call this function explicitly because `Access`\n  implementation would do that.","ref":"Estructura.Aston.html#recalculate_calculated/1"},{"type":"type","title":"Estructura.Aston.t/0","doc":"","ref":"Estructura.Aston.html#t:t/0"},{"type":"function","title":"Estructura.Aston.to_ast/1","doc":"Converts `Estructura.Aston` to the XML AST understandable by `XmlBuilder`","ref":"Estructura.Aston.html#to_ast/1"},{"type":"module","title":"Estructura.Nested","doc":"The nested struct with helpers to easily describe it and produce\n  validation, coercion, and generation helpers.","ref":"Estructura.Nested.html"},{"type":"macro","title":"Estructura.Nested.coerce/1","doc":"DSL helper to produce **`coerce`** callbacks. The syntax is kinda weird,\n  but bear with it, please.\n\nIt’s known to produce warnings in `credo`, I’m working on it.\n```elixir\ncoerce do\n  def data.age(age) when is_float(age), do: {:ok, age}\n  def data.age(age) when is_integer(age), do: {:ok, 1.0 * age}\n  def data.age(age) when is_binary(age), do: {:ok, String.to_float(age)}\n  def data.age(age), do: {:error, \"Could not cast #{inspect(age)} to float\"}\nend\n```","ref":"Estructura.Nested.html#coerce/1"},{"type":"macro","title":"Estructura.Nested.init/1","doc":"","ref":"Estructura.Nested.html#init/1"},{"type":"macro","title":"Estructura.Nested.shape/1","doc":"Declares the shape of the target nested map. the values might be:\n\n- `:string` | `:integer` or another simple `type` understood by\n  [`StreamData`](https://hexdocs.pm/stream_data/StreamData.html#functions)\n- `[type]` to declare a list of elements of a single `type`\n- `map` to declare a nesting level; in such a case, the module with the FQN\n  is created, carrying the struct of the same behaviour\n- `mfa` tuple pointing out to the generator for this value","ref":"Estructura.Nested.html#shape/1"},{"type":"macro","title":"Example - Estructura.Nested.shape/1","doc":"```elixir\ndefmodule User do\n  use Estructura.Nested\n  shape %{\n    name: :string,\n    address: %{city: :string, street: %{name: [:string], house: :string}},\n    data: %{age: :float}\n  }\nend\n\n%User{}\n```\n\nwould result in\n\n```elixir\n%User{\n  address: %User.Address{\n    city: nil,\n    street: %User.Address.Street{house: nil, name: []}\n  },\n  data: %User.Data{age: nil},\n  name: nil\n  }\n```","ref":"Estructura.Nested.html#shape/1-example"},{"type":"macro","title":"Estructura.Nested.validate/1","doc":"DSL helper to produce **`validate`** callbacks. The syntax is kinda weird,\n  but bear with it, please.\n\nIt’s known to produce warnings in `credo`, I’m working on it.\n```elixir\nvalidate do\n  def address.street.postal_code(<<?0, code::binary-size(4)>>),\n    do: {:ok, code}\n  def address.street.postal_code(code),\n    do: {:error, \"Not a postal code (#{inspect(code)})\"}\nend\n```","ref":"Estructura.Nested.html#validate/1"},{"type":"module","title":"Estructura.Nested.Type.UUID","doc":"`Estructura` type for `UUID`","ref":"Estructura.Nested.Type.UUID.html"},{"type":"type","title":"Estructura.Nested.Type.UUID.t/0","doc":"","ref":"Estructura.Nested.Type.UUID.html#t:t/0"},{"type":"protocol","title":"Estructura.Flattenable","doc":"","ref":"Estructura.Flattenable.html"},{"type":"function","title":"Estructura.Flattenable.flatten/2","doc":"The function returning the flattened input.\n\nThis protocol is explicitly handful when deeply nested structs are to be serialized.\n\n```elixir\niex|%_{}|1 ▶ %Estructura.User{}\n%Estructura.User{\n  address: %Estructura.User.Address{\n    city: nil,\n    street: %Estructura.User.Address.Street{house: nil, name: []}\n  },\n  data: %Estructura.User.Data{age: nil},\n  name: nil\n}\n\niex|%_{}|2 ▶ Estructura.Flattenable.flatten(%Estructura.User{}, coupler: \"-\", except: ~w|address-street data-age|)\n%{\"address-city\" => nil, \"birthday\" => nil, \"created_at\" => nil, \"name\" => nil}\niex|%_{}|3 ▶ Estructura.Flattenable.flatten(%Estructura.User{}, only: ~w|address_street data_age|)\n%{\"address_street_house\" => nil, \"data_age\" => nil}\niex|%_{}|4 ▶ Estructura.Flattenable.flatten(%Estructura.User{}, only: ~w|address|)\n%{\"address_city\" => nil, \"address_street_house\" => nil}\n```\n\nAllowed options are:\n\n- **`coupler`** the string to concatenate nested keys with, _default:_ **`-`** \n- **`only`** the list of keys to select\n- **`except`** the list of keys to ignore\n- **`jsonify`** `true` or a json encoder implementation; if set, the values will be jsonified\n\nTo enable it for your struct, use `@derive Estructura.Flattenable` or\n  `@derive {Estructura.Flattenable, options}`. `Estructura` implementations derive it be default.","ref":"Estructura.Flattenable.html#flatten/2"},{"type":"type","title":"Estructura.Flattenable.t/0","doc":"All the types that implement this protocol.","ref":"Estructura.Flattenable.html#t:t/0"},{"type":"protocol","title":"Estructura.Transformer","doc":"","ref":"Estructura.Transformer.html"},{"type":"type","title":"Estructura.Transformer.t/0","doc":"All the types that implement this protocol.","ref":"Estructura.Transformer.html#t:t/0"},{"type":"function","title":"Estructura.Transformer.transform/2","doc":"The function returning the transformed input to be used with `Inspect` protocol.\n\nThis protocol is explicitly handful when deeply nested structs are to be logged/inspected.\n\n```elixir\niex|%_{}|1 ▶ %Estructura.User{}\n%Estructura.User{\n  address: %Estructura.User.Address{\n    city: nil,\n    street: %Estructura.User.Address.Street{house: nil, name: []}\n  },\n  data: %Estructura.User.Data{age: nil},\n  name: nil\n}\n\n\niex|%_{}|2 ▶ Estructura.Transformer.transform %Estructura.User{}, type: false, except: ~w|address.street data.age|\n[address: [city: nil], data: [], name: nil]\niex|%_{}|3 ▶ Estructura.Transformer.transform %Estructura.User{}, type: false, only: ~w|address.street data.age|\n[address: [street: [house: nil, name: []]], data: [age: nil]]\niex|%_{}|4 ▶ Estructura.Transformer.transform %Estructura.User{}, only: ~w|address|\n[\n  *: Estructura.User,\n  address: [\n    *: Estructura.User.Address,\n    city: nil,\n    street: [*: Estructura.User.Address.Street, house: nil, name: []]\n  ]\n]\n```\n\nTo enable it for your struct, use `@derive Estructura.Transformer` or\n  `@derive {Estructura.Transformer, options}`. `Estructura` implementations derive it be default.","ref":"Estructura.Transformer.html#transform/2"},{"type":"module","title":"Estructura.StreamData","doc":"Additional generators to be used with `__generator__/1` generated by `Use Estructura`","ref":"Estructura.StreamData.html"},{"type":"function","title":"Estructura.StreamData.date/1","doc":"Generates an instance of `Date.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#date/1"},{"type":"function","title":"Estructura.StreamData.datetime/1","doc":"Generates an instance of `DateTime.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#datetime/1"},{"type":"function","title":"Estructura.StreamData.id/1","doc":"Identity function as a stream data generator","ref":"Estructura.StreamData.html#id/1"},{"type":"function","title":"Estructura.StreamData.ip4/1","doc":"Generates an instance of `Estructura.Nested.Type.IP.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#ip4/1"},{"type":"function","title":"Estructura.StreamData.ip6/1","doc":"Generates an instance of `Estructura.Nested.Type.IP.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#ip6/1"},{"type":"function","title":"Estructura.StreamData.ip/1","doc":"Generates an instance of `Estructura.Nested.Type.IP.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#ip/1"},{"type":"function","title":"Estructura.StreamData.iterate/2","doc":"Helper to generate _unshrinkable_ streams as per `Stream.iterate/2`.","ref":"Estructura.StreamData.html#iterate/2"},{"type":"opaque","title":"Estructura.StreamData.t/1","doc":"An opaque type that represents an `Estructura.StreamData` generator that generates values\nof type `a`.","ref":"Estructura.StreamData.html#t:t/1"},{"type":"function","title":"Estructura.StreamData.time/1","doc":"Generates an instance of `Time.t()`. This generator is unshrinkable.","ref":"Estructura.StreamData.html#time/1"},{"type":"function","title":"Estructura.StreamData.uri/1","doc":"Generates an instance of `URI.t()`. This generator is unshrinkable.\n\n_See:_ `URI`.\n\n```elixir\n[scheme]://[userinfo]@[host]:[port][path]?[query]#[fragment]\n```","ref":"Estructura.StreamData.html#uri/1"},{"type":"function","title":"Estructura.StreamData.uuid/1","doc":"Generates an instance of `Estructura.Nested.Type.UUID.t()`. This generator is unshrinkable.\n\n**NB** versions `3` and `5` require mandatory `namespace`/`uuid` and `name` options passed.","ref":"Estructura.StreamData.html#uuid/1"},{"type":"module","title":"Estructura.Nested.Type.Enum","doc":"`Estructura` type scaffold for creating enumerated types with a limited set of valid values.\n\nThis module provides a way to create custom types that can only contain predetermined values,\nsimilar to enums in other languages.","ref":"Estructura.Nested.Type.Enum.html"},{"type":"module","title":"Usage - Estructura.Nested.Type.Enum","doc":"There are two ways to use this type scaffold:","ref":"Estructura.Nested.Type.Enum.html#module-usage"},{"type":"module","title":"1. Using the module directly - Estructura.Nested.Type.Enum","doc":"iex> require Estructura.Nested.Type.Enum\n    ...> Estructura.Nested.Type.Enum.type_module_ast(Status, [:pending, :active, :completed])\n    ...> apply(Status, :validate, [:pending])\n    {:ok, :pending}\n    iex> apply(Status, :validate, [:invalid])\n    {:error, \"Expected :invalid to be one of: [:pending, :active, :completed]\"}","ref":"Estructura.Nested.Type.Enum.html#module-1-using-the-module-directly"},{"type":"module","title":"2. Using the use macro - Estructura.Nested.Type.Enum","doc":"iex> defmodule Role do\n    ...>   use Estructura.Nested.Type.Enum, elements: [:admin, :user, :guest]\n    ...> end\n    ...> Role.validate(:admin)\n    {:ok, :admin}","ref":"Estructura.Nested.Type.Enum.html#module-2-using-the-use-macro"},{"type":"module","title":"Configuration Options - Estructura.Nested.Type.Enum","doc":"The scaffold accepts the following options:\n\n- `:elements` - (required) List of valid values for the enum\n- `:coercer` - (optional) Function to coerce input values\n- `:encoder` - (optional) Function to encode values for JSON","ref":"Estructura.Nested.Type.Enum.html#module-configuration-options"},{"type":"module","title":"Example with Custom Coercion - Estructura.Nested.Type.Enum","doc":"defmodule Status do\n      use Estructura.Nested.Type.Enum,\n        elements: [:pending, :active, :completed],\n        coercer: fn\n          str when is_binary(str) -> {:ok, String.to_existing_atom(str)}\n          atom when is_atom(atom) -> {:ok, atom}\n          other -> {:error, \"Cannot coerce #{inspect(other)} to status\"}\n        end\n    end","ref":"Estructura.Nested.Type.Enum.html#module-example-with-custom-coercion"},{"type":"module","title":"Example with Custom JSON Encoding - Estructura.Nested.Type.Enum","doc":"defmodule Role do\n      use Estructura.Nested.Type.Enum,\n        elements: [:admin, :user, :guest],\n        encoder: fn role, opts -> Jason.Encode.string(Atom.to_string(role), opts) end\n    end","ref":"Estructura.Nested.Type.Enum.html#module-example-with-custom-json-encoding"},{"type":"module","title":"Generated Functions - Estructura.Nested.Type.Enum","doc":"The scaffold implements the `Estructura.Nested.Type` behaviour and provides:\n\n- `generate/1` - Generates random values from the enum for testing\n- `coerce/1` - Attempts to convert input into a valid enum value\n- `validate/1` - Ensures a value is part of the enum","ref":"Estructura.Nested.Type.Enum.html#module-generated-functions"},{"type":"module","title":"Generation Options - Estructura.Nested.Type.Enum","doc":"The `generate/1` function accepts:\n- `:only` - List of elements to generate from (default: all elements)\n- `:except` - List of elements to exclude from generation\n\n```elixir\nRole.generate(only: [:admin, :user]) |> Enum.take(1) |> List.first()\n#⇒ :admin # or :user\n\nRole.generate(except: [:guest]) |> Enum.take(1) |> List.first()\n#⇒ :admin # or :user\n```","ref":"Estructura.Nested.Type.Enum.html#module-generation-options"},{"type":"macro","title":"Estructura.Nested.Type.Enum.__using__/1","doc":"Implements the enum type directly in the current module.","ref":"Estructura.Nested.Type.Enum.html#__using__/1"},{"type":"macro","title":"Options - Estructura.Nested.Type.Enum.__using__/1","doc":"See the module documentation for available options.","ref":"Estructura.Nested.Type.Enum.html#__using__/1-options"},{"type":"macro","title":"Examples - Estructura.Nested.Type.Enum.__using__/1","doc":"defmodule Role do\n      use Estructura.Nested.Type.Enum, elements: [:admin, :user, :guest]\n    end","ref":"Estructura.Nested.Type.Enum.html#__using__/1-examples"},{"type":"function","title":"Estructura.Nested.Type.Enum.type_module_ast/2","doc":"Creates a new enum type module with the given name and options.","ref":"Estructura.Nested.Type.Enum.html#type_module_ast/2"},{"type":"function","title":"Options - Estructura.Nested.Type.Enum.type_module_ast/2","doc":"See the module documentation for available options.","ref":"Estructura.Nested.Type.Enum.html#type_module_ast/2-options"},{"type":"function","title":"Examples - Estructura.Nested.Type.Enum.type_module_ast/2","doc":"```elixir\nEstructura.Nested.Type.Enum.type_module_ast(__MODULE__, elements: [:pending, :active])\n```","ref":"Estructura.Nested.Type.Enum.html#type_module_ast/2-examples"},{"type":"module","title":"Estructura.Nested.Type.Tags","doc":"`Estructura` type for a set of predefined values, might be used as an implementation generator.","ref":"Estructura.Nested.Type.Tags.html"},{"type":"behaviour","title":"Estructura.Nested.Type","doc":"The type to be used for coercing, validating, and generation\n  of the implementation’s instances.","ref":"Estructura.Nested.Type.html"},{"type":"callback","title":"Estructura.Nested.Type.coerce/1","doc":"Coerces the value coming from outside","ref":"Estructura.Nested.Type.html#c:coerce/1"},{"type":"callback","title":"Estructura.Nested.Type.generate/0","doc":"The generator for the type","ref":"Estructura.Nested.Type.html#c:generate/0"},{"type":"callback","title":"Estructura.Nested.Type.generate/1","doc":"The generator for the type accepting options","ref":"Estructura.Nested.Type.html#c:generate/1"},{"type":"callback","title":"Estructura.Nested.Type.validate/1","doc":"Validates the value as being correct","ref":"Estructura.Nested.Type.html#c:validate/1"},{"type":"module","title":"Estructura.Nested.Type.Date","doc":"`Estructura` type implementation for handling `Date` values.\n\nThis type provides functionality for:\n- Generating Date values for testing\n- Coercing various inputs into Date format\n- Validating Date values","ref":"Estructura.Nested.Type.Date.html"},{"type":"module","title":"Examples - Estructura.Nested.Type.Date","doc":"iex> alias Estructura.Nested.Type.Date\n    iex> Date.validate(~D[2024-01-01])\n    {:ok, ~D[2024-01-01]}\n\n    iex> alias Estructura.Nested.Type.Date\n    iex> Date.validate(\"not a date\")\n    {:error, \"Expected date, got: \\\"not a date\\\"\"}\n\nThe type implements the `Estructura.Nested.Type` behaviour, providing:\n- `generate/1` - Creates random Date values for property testing\n- `coerce/1` - Attempts to convert input into a Date\n- `validate/1` - Ensures a value is a valid Date","ref":"Estructura.Nested.Type.Date.html#module-examples"},{"type":"function","title":"Estructura.Nested.Type.Date.coerce/1","doc":"Attempts to coerce a value into a Date.\n\nDelegates to `Estructura.Coercers.Date.coerce/1` which handles various input formats\nincluding strings in ISO 8601 format and maps with date components.","ref":"Estructura.Nested.Type.Date.html#coerce/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.Date.coerce/1","doc":"iex> Date.coerce(\"2024-01-01\")\n    {:ok, ~D[2024-01-01]}\n\n    iex> Date.coerce(%{year: 2024, month: 1, day: 1})\n    {:ok, ~D[2024-01-01]}\n\n    iex> Date.coerce(\"invalid\")\n    {:error, \"Invalid Date format\"}","ref":"Estructura.Nested.Type.Date.html#coerce/1-examples"},{"type":"function","title":"Estructura.Nested.Type.Date.generate/1","doc":"Generates random Date values for property-based testing.","ref":"Estructura.Nested.Type.Date.html#generate/1"},{"type":"function","title":"Options - Estructura.Nested.Type.Date.generate/1","doc":"Accepts all options supported by `Estructura.StreamData.date/1`.","ref":"Estructura.Nested.Type.Date.html#generate/1-options"},{"type":"function","title":"Examples - Estructura.Nested.Type.Date.generate/1","doc":"iex> Date.generate() |> Enum.take(1) |> List.first()\n    #Date<...>\n\n    iex> Date.generate(from: ~D[2024-01-01], to: ~D[2024-12-31]) |> Enum.take(1) |> List.first()\n    #Date<2024-...>","ref":"Estructura.Nested.Type.Date.html#generate/1-examples"},{"type":"function","title":"Estructura.Nested.Type.Date.validate/1","doc":"Validates that a term is a valid Date.\n\nReturns `{:ok, date}` for valid Date values,\nor `{:error, reason}` for invalid ones.","ref":"Estructura.Nested.Type.Date.html#validate/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.Date.validate/1","doc":"iex> Date.validate(~D[2024-01-01])\n    {:ok, ~D[2024-01-01]}\n\n    iex> Date.validate(\"2024\")\n    {:error, \"Expected date, got: \\\"2024\\\"\"}","ref":"Estructura.Nested.Type.Date.html#validate/1-examples"},{"type":"module","title":"Estructura.Nested.Type.DateTime","doc":"`Estructura` type implementation for handling `DateTime` values.\n\nThis type provides functionality for:\n- Generating DateTime values for testing\n- Coercing various inputs into DateTime format\n- Validating DateTime values","ref":"Estructura.Nested.Type.DateTime.html"},{"type":"module","title":"Examples - Estructura.Nested.Type.DateTime","doc":"iex> alias Estructura.Nested.Type.DateTime\n    iex> DateTime.validate(~U[2024-01-01 10:00:00Z])\n    {:ok, ~U[2024-01-01 10:00:00Z]}\n\n    iex> alias Estructura.Nested.Type.DateTime\n    iex> DateTime.validate(\"not a datetime\")\n    {:error, \"Expected date, got: \\\"not a datetime\\\"\"}\n\nThe type implements the `Estructura.Nested.Type` behaviour, providing:\n- `generate/1` - Creates random DateTime values for property testing\n- `coerce/1` - Attempts to convert input into a DateTime\n- `validate/1` - Ensures a value is a valid DateTime","ref":"Estructura.Nested.Type.DateTime.html#module-examples"},{"type":"function","title":"Estructura.Nested.Type.DateTime.coerce/1","doc":"Attempts to coerce a value into a DateTime.\n\nDelegates to `Estructura.Coercers.DateTime.coerce/1` which handles various input formats.","ref":"Estructura.Nested.Type.DateTime.html#coerce/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.DateTime.coerce/1","doc":"iex> DateTime.coerce(\"2024-01-01T10:00:00Z\")\n    {:ok, ~U[2024-01-01 10:00:00Z]}\n\n    iex> DateTime.coerce(\"invalid\")\n    {:error, \"Invalid DateTime format\"}","ref":"Estructura.Nested.Type.DateTime.html#coerce/1-examples"},{"type":"function","title":"Estructura.Nested.Type.DateTime.generate/1","doc":"Generates random DateTime values for property-based testing.","ref":"Estructura.Nested.Type.DateTime.html#generate/1"},{"type":"function","title":"Options - Estructura.Nested.Type.DateTime.generate/1","doc":"Accepts all options supported by `Estructura.StreamData.datetime/1`.","ref":"Estructura.Nested.Type.DateTime.html#generate/1-options"},{"type":"function","title":"Examples - Estructura.Nested.Type.DateTime.generate/1","doc":"iex> DateTime.generate() |> Enum.take(1) |> List.first()\n    #DateTime<...>","ref":"Estructura.Nested.Type.DateTime.html#generate/1-examples"},{"type":"function","title":"Estructura.Nested.Type.DateTime.validate/1","doc":"Validates that a term is a valid DateTime.\n\nReturns `{:ok, datetime}` for valid DateTime values,\nor `{:error, reason}` for invalid ones.","ref":"Estructura.Nested.Type.DateTime.html#validate/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.DateTime.validate/1","doc":"iex> DateTime.validate(~U[2024-01-01 10:00:00Z])\n    {:ok, ~U[2024-01-01 10:00:00Z]}\n\n    iex> DateTime.validate(\"2024\")\n    {:error, \"Expected date, got: \\\"2024\\\"\"}","ref":"Estructura.Nested.Type.DateTime.html#validate/1-examples"},{"type":"module","title":"Estructura.Nested.Type.IP","doc":"`Estructura` type for `Date`","ref":"Estructura.Nested.Type.IP.html"},{"type":"macro","title":"Estructura.Nested.Type.IP.sigil_IP/2","doc":"","ref":"Estructura.Nested.Type.IP.html#sigil_IP/2"},{"type":"type","title":"Estructura.Nested.Type.IP.t/0","doc":"","ref":"Estructura.Nested.Type.IP.html#t:t/0"},{"type":"module","title":"Estructura.Nested.Type.String","doc":"`Estructura` type for `String`","ref":"Estructura.Nested.Type.String.html"},{"type":"module","title":"Estructura.Nested.Type.Time","doc":"`Estructura` type implementation for handling `Time` values.\n\nThis type provides functionality for:\n- Generating Time values for testing\n- Coercing various inputs into Time format\n- Validating Time values","ref":"Estructura.Nested.Type.Time.html"},{"type":"module","title":"Examples - Estructura.Nested.Type.Time","doc":"iex> alias Estructura.Nested.Type.Time\n    iex> Time.validate(~T[10:00:00])\n    {:ok, ~T[10:00:00]}\n\n    iex> alias Estructura.Nested.Type.Time\n    iex> Time.validate(\"not a time\")\n    {:error, \"Expected time, got: \\\"not a time\\\"\"}\n\nThe type implements the `Estructura.Nested.Type` behaviour, providing:\n- `generate/1` - Creates random Time values for property testing\n- `coerce/1` - Attempts to convert input into a Time\n- `validate/1` - Ensures a value is a valid Time","ref":"Estructura.Nested.Type.Time.html#module-examples"},{"type":"function","title":"Estructura.Nested.Type.Time.coerce/1","doc":"Attempts to coerce a value into a Time.\n\nDelegates to `Estructura.Coercers.Time.coerce/1` which handles various input formats\nincluding strings in ISO 8601 format and maps with time components.","ref":"Estructura.Nested.Type.Time.html#coerce/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.Time.coerce/1","doc":"iex> Time.coerce(\"10:00:00\")\n    {:ok, ~T[10:00:00]}\n\n    iex> Time.coerce(%{hour: 10, minute: 0, second: 0})\n    {:ok, ~T[10:00:00]}\n\n    iex> Time.coerce(\"invalid\")\n    {:error, \"Invalid Time format\"}","ref":"Estructura.Nested.Type.Time.html#coerce/1-examples"},{"type":"function","title":"Estructura.Nested.Type.Time.generate/1","doc":"Generates random Time values for property-based testing.","ref":"Estructura.Nested.Type.Time.html#generate/1"},{"type":"function","title":"Options - Estructura.Nested.Type.Time.generate/1","doc":"Accepts all options supported by `Estructura.StreamData.time/1`.","ref":"Estructura.Nested.Type.Time.html#generate/1-options"},{"type":"function","title":"Examples - Estructura.Nested.Type.Time.generate/1","doc":"iex> Time.generate() |> Enum.take(1) |> List.first()\n    #Time<...>\n\n    iex> Time.generate(from: ~T[09:00:00], to: ~T[17:00:00]) |> Enum.take(1) |> List.first()\n    #Time<...>","ref":"Estructura.Nested.Type.Time.html#generate/1-examples"},{"type":"function","title":"Estructura.Nested.Type.Time.validate/1","doc":"Validates that a term is a valid Time.\n\nReturns `{:ok, time}` for valid Time values,\nor `{:error, reason}` for invalid ones.","ref":"Estructura.Nested.Type.Time.html#validate/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.Time.validate/1","doc":"iex> Time.validate(~T[10:00:00])\n    {:ok, ~T[10:00:00]}\n\n    iex> Time.validate(\"10:00\")\n    {:error, \"Expected time, got: \\\"10:00\\\"\"}","ref":"Estructura.Nested.Type.Time.html#validate/1-examples"},{"type":"module","title":"Estructura.Nested.Type.URI","doc":"`Estructura` type implementation for handling `URI` values.\n\nThis type provides functionality for:\n- Generating URI values for testing\n- Coercing string and map inputs into URI structs\n- Validating URI values","ref":"Estructura.Nested.Type.URI.html"},{"type":"module","title":"Examples - Estructura.Nested.Type.URI","doc":"Estructura.Nested.Type.URI.validate(URI.parse(\"https://example.com/path?query=value\"))\n#⇒ {:ok, %URI{scheme: \"https\", host: \"example.com\", path: \"/path\", query: \"query=value\"}}\n\nEstructura.Nested.Type.URI.validate(\"not a uri\")\n#⇒ {:error, \"Expected URI, got: \\\"not a uri\\\"\"}\n\nThe type implements the `Estructura.Nested.Type` behaviour, providing:\n- `generate/1` - Creates random URI values for property testing\n- `coerce/1` - Attempts to convert input into a URI\n- `validate/1` - Ensures a value is a valid URI","ref":"Estructura.Nested.Type.URI.html#module-examples"},{"type":"function","title":"Estructura.Nested.Type.URI.coerce/1","doc":"Attempts to coerce a value into a URI.\n\nDelegates to `URI.new/1` which handles various input formats including:\n- URI strings (\"https://example.com\")\n- Maps with URI components","ref":"Estructura.Nested.Type.URI.html#coerce/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.URI.coerce/1","doc":"Estructura.Nested.Type.URI.coerce(\"https://example.com/path?query=value\")\n#⇒ {:ok, %URI{scheme: \"https\", host: \"example.com\", path: \"/path\", query: \"query=value\"}}\n\nEstructura.Nested.Type.URI.coerce(\"invalid uri\")\n#⇒ {:error, \"Invalid URI format\"}","ref":"Estructura.Nested.Type.URI.html#coerce/1-examples"},{"type":"function","title":"Estructura.Nested.Type.URI.generate/1","doc":"Generates random URI values for property-based testing.","ref":"Estructura.Nested.Type.URI.html#generate/1"},{"type":"function","title":"Options - Estructura.Nested.Type.URI.generate/1","doc":"Accepts all options supported by `Estructura.StreamData.uri/1`, including:\n- `:schemes` - List of allowed schemes (default: [\"http\", \"https\"])\n- `:hosts` - List of allowed hosts (default: generates random hosts)\n- `:paths` - List of allowed paths (default: generates random paths)\n- `:with_query` - Whether to include query parameters (default: true)","ref":"Estructura.Nested.Type.URI.html#generate/1-options"},{"type":"function","title":"Examples - Estructura.Nested.Type.URI.generate/1","doc":"Estructura.Nested.Type.URI.generate() |> Enum.take(1) |> List.first()\n#⇒ %URI{scheme: \"https\", host: \"example.com\", path: \"/some/path\"}\n\nEstructura.Nested.Type.URI.generate(schemes: [\"ftp\"]) |> Enum.take(1) |> List.first()\n#⇒ %URI{scheme: \"ftp\", host: \"example.com\", path: \"/\"}","ref":"Estructura.Nested.Type.URI.html#generate/1-examples"},{"type":"function","title":"Estructura.Nested.Type.URI.validate/1","doc":"Validates that a term is a valid URI.\n\nReturns `{:ok, uri}` for valid URI values,\nor `{:error, reason}` for invalid ones.","ref":"Estructura.Nested.Type.URI.html#validate/1"},{"type":"function","title":"Examples - Estructura.Nested.Type.URI.validate/1","doc":"Estructura.Nested.Type.URI.validate(URI.parse(\"https://example.com\"))\n#⇒ {:ok, %URI{scheme: \"https\", host: \"example.com\"}}\n\nEstructura.Nested.Type.URI.validate(\"not a uri\")\n#⇒ {:error, \"Expected URI, got: \\\"not a uri\\\"\"}","ref":"Estructura.Nested.Type.URI.html#validate/1-examples"},{"type":"module","title":"Estructura.Lazy","doc":"The field stub allowing lazy instantiation of `Estructura` fields.","ref":"Estructura.Lazy.html"},{"type":"function","title":"Estructura.Lazy.apply/3","doc":"Apply the lazy getter to the data passed as an argument","ref":"Estructura.Lazy.html#apply/3"},{"type":"function","title":"Examples - Estructura.Lazy.apply/3","doc":"iex> lazy = Estructura.Lazy.new(&System.fetch_env/1)\n    ...> Estructura.Lazy.apply(lazy, \"LANG\").value\n    System.fetch_env(\"LANG\")","ref":"Estructura.Lazy.html#apply/3-examples"},{"type":"type","title":"Estructura.Lazy.cached/0","doc":"","ref":"Estructura.Lazy.html#t:cached/0"},{"type":"type","title":"Estructura.Lazy.getter/0","doc":"","ref":"Estructura.Lazy.html#t:getter/0"},{"type":"type","title":"Estructura.Lazy.key/0","doc":"","ref":"Estructura.Lazy.html#t:key/0"},{"type":"function","title":"Estructura.Lazy.new/2","doc":"Create the new struct with the getter passed as an argument","ref":"Estructura.Lazy.html#new/2"},{"type":"function","title":"Estructura.Lazy.stale?/1","doc":"Validates if the value is not stale yet according to `expires_in` setting","ref":"Estructura.Lazy.html#stale?/1"},{"type":"type","title":"Estructura.Lazy.t/0","doc":"","ref":"Estructura.Lazy.html#t:t/0"},{"type":"type","title":"Estructura.Lazy.value/0","doc":"","ref":"Estructura.Lazy.html#t:value/0"},{"type":"module","title":"Estructura.LazyMap","doc":"The implementation of lazy map implementing lazy `Access` for its keys.\n\n`Estructura.LazyMap` is backed by the “raw” object and a key-value pairs\n  where values might be instances of `Estructura.Lazy`. If this is a case,\n  they will be accessed through `Lazy` implementation.\n\nValues might be also raw values, which makes `LazyMap` a drop-in replacement\n  of standard _Elixir_ maps, assuming they are accessed through `Access`\n  only (e. g. `map[:key]` and not `map.key`.)","ref":"Estructura.LazyMap.html"},{"type":"function","title":"Estructura.LazyMap.fetch_all/1","doc":"Eagerly instantiates the data","ref":"Estructura.LazyMap.html#fetch_all/1"},{"type":"function","title":"Estructura.LazyMap.keys/1","doc":"Returns all the keys of the underlying map","ref":"Estructura.LazyMap.html#keys/1"},{"type":"function","title":"Estructura.LazyMap.new/2","doc":"Creates new instance of `LazyMap` with a second parameter being a backed up object,\n  which would be used for lazy retrieving data for values, when value is an instance\n  of `Estructura.Lazy`.","ref":"Estructura.LazyMap.html#new/2"},{"type":"function","title":"Examples - Estructura.LazyMap.new/2","doc":"iex> lm = Estructura.LazyMap.new(\n    ...>   [int: Estructura.Lazy.new(&Estructura.LazyInst.parse_int/1)], \"42\")\n    ...> get_in lm, [:int]\n    42","ref":"Estructura.LazyMap.html#new/2-examples"},{"type":"type","title":"Estructura.LazyMap.t/0","doc":"","ref":"Estructura.LazyMap.html#t:t/0"},{"type":"behaviour","title":"Estructura.Coercer","doc":"Behaviour for coercion delegates. Instead of implementing the coercion handlers\n  in `Estructura.Nested` inplace, one might do\n    \n```elixir\ncoerce do\n  defdelegate foo.bar.created_at(value), to: :date\nend\n```","ref":"Estructura.Coercer.html"},{"type":"callback","title":"Estructura.Coercer.coerce/1","doc":"","ref":"Estructura.Coercer.html#c:coerce/1"},{"type":"module","title":"Estructura.Coercers.Atom","doc":"Coercer for `:atom`, coercing strings by a call to `String.to_existing_atom/1`","ref":"Estructura.Coercers.Atom.html"},{"type":"module","title":"Estructura.Coercers.Date","doc":"Default coercer for `:date`, coercing strings (_ISO8601_) and integers (_epoch_)","ref":"Estructura.Coercers.Date.html"},{"type":"module","title":"Estructura.Coercers.DateTime","doc":"Default coercer for `:datetime`, coercing strings (_ISO8601_) and integers (_epoch_)","ref":"Estructura.Coercers.DateTime.html"},{"type":"module","title":"Estructura.Coercers.Datetime","doc":"Default coercer for `:datetime`, coercing strings (_ISO8601_) and integers (_epoch_)","ref":"Estructura.Coercers.Datetime.html"},{"type":"module","title":"Estructura.Coercers.Float","doc":"Default coercer for `:float`, coercing strings and integers by multiplying by `1.0`","ref":"Estructura.Coercers.Float.html"},{"type":"module","title":"Estructura.Coercers.Integer","doc":"Default coercer for `:integer`, coercing strings and floats by rounding","ref":"Estructura.Coercers.Integer.html"},{"type":"module","title":"Estructura.Coercers.NullableDate","doc":"Nullable coercer for `:date`, coercing strings and floats by rounding, allows `nil` value","ref":"Estructura.Coercers.NullableDate.html"},{"type":"module","title":"Estructura.Coercers.NullableDatetime","doc":"Nullable coercer for `:datetime`, coercing strings and floats by rounding, allows `nil` value","ref":"Estructura.Coercers.NullableDatetime.html"},{"type":"module","title":"Estructura.Coercers.NullableFloat","doc":"Nullable coercer for `:float`, coercing strings and floats by rounding, allows `nil` value","ref":"Estructura.Coercers.NullableFloat.html"},{"type":"module","title":"Estructura.Coercers.NullableInteger","doc":"Nullable coercer for `:integer`, coercing strings and floats by rounding, allows `nil` value","ref":"Estructura.Coercers.NullableInteger.html"},{"type":"module","title":"Estructura.Coercers.NullableTime","doc":"Nullable coercer for `:time`, coercing strings and floats by rounding, allows `nil` value","ref":"Estructura.Coercers.NullableTime.html"},{"type":"module","title":"Estructura.Coercers.Time","doc":"Default coercer for `:time`, coercing strings (_ISO8601_) and integers (_epoch_)","ref":"Estructura.Coercers.Time.html"},{"type":"module","title":"Estructura.Config","doc":"The configuration of the `Estructura` that is built from the parameters,\npassed as a second argument in a call to `use Estructura`.\n\nOne normally does not need to meddle with this module.","ref":"Estructura.Config.html"},{"type":"type","title":"Estructura.Config.generator/0","doc":"The generator to be passed to `use Estructura` should be given in one of these forms","ref":"Estructura.Config.html#t:generator/0"},{"type":"type","title":"Estructura.Config.key/0","doc":"The structure key","ref":"Estructura.Config.html#t:key/0"},{"type":"type","title":"Estructura.Config.t/0","doc":"The configuration of `Estructura`","ref":"Estructura.Config.html#t:t/0"},{"type":"type","title":"Estructura.Config.value/0","doc":"The structure value","ref":"Estructura.Config.html#t:value/0"},{"type":"module","title":"Estructura.Full","doc":"Full Example","ref":"Estructura.Full.html"},{"type":"function","title":"Estructura.Full.__generator__/0","doc":"See `Estructura.Full.__generator__/1`","ref":"Estructura.Full.html#__generator__/0"},{"type":"function","title":"Estructura.Full.__generator__/1","doc":"Returns the generator to be used in `StreamData`-powered property testing, based\n  on the specification given to `use Estructura`, which contained","ref":"Estructura.Full.html#__generator__/1"},{"type":"function","title":"shape - Estructura.Full.__generator__/1","doc":"```elixir\n%Estructura.Config{\n  access: true,\n  coercion: [:foo],\n  validation: true,\n  calculated: [],\n  collectable: :bar,\n  enumerable: true,\n  generator: [\n    foo: {StreamData, :integer, [0..1000]},\n    bar: {StreamData, :string, [:alphanumeric]},\n    baz: {StreamData, :fixed_map,\n     [[key1: {StreamData, :integer}, key2: {StreamData, :integer}]]},\n    zzz: &Estructura.Full.zzz_generator/0\n  ]\n}\n```\n\n\nThe argument given would be used as a template to generate new values.","ref":"Estructura.Full.html#__generator__/1-shape"},{"type":"function","title":"Estructura.Full.get/3","doc":"Gets the value for the given key from the structure","ref":"Estructura.Full.html#get/3"},{"type":"function","title":"Estructura.Full.put/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns `{:ok, updated_struct}` or `{:error, reason}` if there is no such key","ref":"Estructura.Full.html#put/3"},{"type":"function","title":"Estructura.Full.put!/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns the value or raises if there is no such key","ref":"Estructura.Full.html#put!/3"},{"type":"function","title":"Estructura.Full.recalculate_calculated/1","doc":"Recalculates calculated fields for the instance of Estructura.Full.\n\nNormally one would not need to call this function explicitly because `Access`\n  implementation would do that.","ref":"Estructura.Full.html#recalculate_calculated/1"},{"type":"module","title":"Estructura.User","doc":"Nested example. The source code of the file follows.\n\n```elixir\n\ndefmodule Calculated do\n  @moduledoc false\n  def person(this) do\n    this.name <> \", \" <> this.address.city\n  end\nend\n\nuse Estructura.Nested, calculated: [person: &Calculated.person/1]\n\nshape %{\n  created_at: :datetime,\n  name: {:string, kind_of_codepoints: Enum.concat([?a..?c, ?l..?o])},\n  address: %{city: :string, street: %{name: [:string], house: :string}},\n  person: :string,\n  homepage: {:list_of, Estructura.Nested.Type.URI},\n  ip: Estructura.Nested.Type.IP,\n  data: %{age: :float},\n  birthday: Estructura.Nested.Type.Date,\n  title: {Estructura.Nested.Type.Enum, ~w|junior middle señor|},\n  tags: {:tags, ~w|backend frontend|}\n}\n\ninit %{\n  name: \"Aleksei\",\n  address: %{city: \"Barcelona\"}\n}\n\ncoerce do\n  def data.age(age) when is_float(age), do: {:ok, age}\n  def data.age(age) when is_integer(age), do: {:ok, 1.0 * age}\n  def data.age(age) when is_binary(age) do\n    age\n    |> Float.parse()\n    |> case do\n      {age, \"\"} -> {:ok, age}\n      {age, _rest} -> {:ok, age}\n      :error -> {:ok, 0.0}\n    end\n  end\nend\n\ncoerce do\n  def name(value) when is_binary(value), do: {:ok, value}\n  def name(value) when is_atom(value), do: {:ok, Atom.to_string(value)}\nend\n\nvalidate do\n  def address.street.house(house), do: {:ok, house}\nend\n```\n\nNow one can cast it from map as below\n\n```elixir\nUser.cast %{\n  address: %{city: \"London\", street: %{name: \"Baker\", house: \"221 Bis\"}},\n  data: %{age: 32},\n  name: \"Watson\",\n  birthday: \"1973-09-30\",\n  title: \"señor\",\n  tags: [\"backend\"]}\n\n{:ok,\n   %Estructura.User{\n     address: %Estructura.User.Address{\n       city: \"London\",\n       street: %Estructura.User.Address.Street{house: \"221 Bis\", name: \"Baker\"}\n     },\n     data: %Estructura.User.Data{age: 32.0},\n     name: \"Watson\",\n     person: \"Watson, London\",\n     birthday: ~D[1973-09-30],\n     title: \"señor\",\n     tags: [\"backend\"]}\n```","ref":"Estructura.User.html"},{"type":"function","title":"Estructura.User.__generator__/0","doc":"See `Estructura.User.__generator__/1`","ref":"Estructura.User.html#__generator__/0"},{"type":"function","title":"Estructura.User.__generator__/1","doc":"Returns the generator to be used in `StreamData`-powered property testing, based\n  on the specification given to `use Estructura.Nested`, which contained","ref":"Estructura.User.html#__generator__/1"},{"type":"function","title":"shape - Estructura.User.__generator__/1","doc":"```elixir\n%{\n  data: %{age: :float},\n  name: {:string, [kind_of_codepoints: ~c\"abclmno\"]},\n  address: %{\n    city: :string,\n    street: %{name: [:string], house: :positive_integer}\n  },\n  ip: Estructura.Nested.Type.IP,\n  title: {Estructura.Nested.Type.Enum, [\"junior\", \"middle\", \"señor\"]},\n  tags: {Estructura.Nested.Type.Tags, [\"backend\", \"frontend\"]},\n  person: :string,\n  created_at: :datetime,\n  homepage: {:list_of, Estructura.Nested.Type.URI},\n  birthday: Estructura.Nested.Type.Date\n}\n```","ref":"Estructura.User.html#__generator__/1-shape"},{"type":"function","title":"coerce - Estructura.User.__generator__/1","doc":"```elixir\n  def coerce_age(age) when is_float(age) do\n    {:ok, age}\n  end\n  def coerce_age(age) when is_integer(age) do\n    {:ok, 1.0 * age}\n  end\n  def coerce_age(age) when is_binary(age) do\n    age\n    |> Float.parse()\n    |> case do\n      {age, \"\"} -> {:ok, age}\n      {age, _rest} -> {:ok, age}\n      :error -> {:ok, 0.0}\n    end\n  end\n  def coerce_created_at(value) do\n    Estructura.Coercers.Datetime.coerce(value)\n  end\n  def coerce_homepage(value) when is_list(value) do\n    value\n    |> Enum.reduce({:ok, []}, fn\n      value, {:error, errors} ->\n        case Estructura.Nested.Type.URI.coerce(value) do\n          {:ok, _} -> {:error, errors}\n          {:error, error} -> {:error, [error | errors]}\n        end\n\n      value, {:ok, result} ->\n        case Estructura.Nested.Type.URI.coerce(value) do\n          {:ok, coerced} -> {:ok, [coerced | result]}\n          {:error, error} -> {:error, [error]}\n        end\n    end)\n    |> then(fn {kind, list} -> {kind, Enum.reverse(list)} end)\n  end\n```","ref":"Estructura.User.html#__generator__/1-coerce"},{"type":"function","title":"validate - Estructura.User.__generator__/1","doc":"```elixir\n  def validate_house(house) do\n    {:ok, house}\n  end\n  def validate_age(age) when age > 0 do\n    {:ok, age}\n  end\n  def validate_age(age) do\n    {:error, \"Age must be positive, given: #{age}\"}\n  end\n```\n\n\nThe argument given would be used as a template to generate new values.","ref":"Estructura.User.html#__generator__/1-validate"},{"type":"function","title":"Estructura.User.cast/2","doc":"Casts the map representation as given to `Estructura.Nested.shape/1` to\n  the nested `Estructura` instance.\n\nIf `split: true` is passed as an option, it will attempt to put `foo_bar` into nested `%{foo: %{bar: _}}`","ref":"Estructura.User.html#cast/2"},{"type":"function","title":"Estructura.User.cast!/2","doc":"Same as `cast/2` but raises on errors.","ref":"Estructura.User.html#cast!/2"},{"type":"function","title":"Estructura.User.get/3","doc":"Gets the value for the given key from the structure","ref":"Estructura.User.html#get/3"},{"type":"function","title":"Estructura.User.parse/1","doc":"Safely parses the json, applying all the specified validations and coercions","ref":"Estructura.User.html#parse/1"},{"type":"function","title":"Estructura.User.parse!/1","doc":"Same as `parse/1` but either returns the result of successful parsing or raises","ref":"Estructura.User.html#parse!/1"},{"type":"function","title":"Estructura.User.put/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns `{:ok, updated_struct}` or `{:error, reason}` if there is no such key","ref":"Estructura.User.html#put/3"},{"type":"function","title":"Estructura.User.put!/3","doc":"Puts the value for the given key into the structure, passing coercion _and_ validation,\n  returns the value or raises if there is no such key","ref":"Estructura.User.html#put!/3"},{"type":"function","title":"Estructura.User.recalculate_calculated/1","doc":"Recalculates calculated fields for the instance of Estructura.User.\n\nNormally one would not need to call this function explicitly because `Access`\n  implementation would do that.","ref":"Estructura.User.html#recalculate_calculated/1"},{"type":"type","title":"Estructura.User.t/0","doc":"","ref":"Estructura.User.html#t:t/0"},{"type":"function","title":"Estructura.User.validate/1","doc":"Validates the instance of `Estructura.User` with all the defined validators.","ref":"Estructura.User.html#validate/1"},{"type":"extras","title":"Estructura","doc":"# Estructura    [![Kantox ❤ OSS](https://img.shields.io/badge/❤-kantox_oss-informational.svg)](https://kantox.com/)  [![Test](https://github.com/am-kantox/estructura/workflows/Test/badge.svg)](https://github.com/am-kantox/estructura/actions?query=workflow%3ATest)  [![Dialyzer](https://github.com/am-kantox/estructura/workflows/Dialyzer/badge.svg)](https://github.com/am-kantox/estructura/actions?query=workflow%3ADialyzer)\n\n**Extensions for _Elixir_ structures.**","ref":"readme.html"},{"type":"extras","title":"Installation - Estructura","doc":"```elixir\ndef deps do\n  [\n    {:estructura, \"~> 0.1\"},\n    # optionally you might want to add `boundary` library \n    # it is used by `estructura` and many other projects\n    # more info: https://hexdocs.pm/boundary\n    {:boundary, \"~> 0.9\", runtime: false}\n  ]\nend\n```\nI suggest adding [`boundary`](https://hexdocs.pm/boundary) as a dependency since that is used in this project.","ref":"readme.html#installation"},{"type":"extras","title":"Features - Estructura","doc":"","ref":"readme.html#features"},{"type":"extras","title":"Nested Structures - Estructura","doc":"`Estructura.Nested` provides powerful nested structure support with validation, coercion, and generation capabilities:\n\n```elixir\ndefmodule User do\n  use Estructura.Nested\n\n  defstruct [\n    name: \"\",\n    address: %{\n      city: \"\",\n      street: %{name: \"\", house: 0}\n    }\n  ]\n\n  # Validation rules\n  def validate(:name, value), do: String.length(value) > 0\n  def validate(\"address.street.house\", value), do: value > 0\nend\n\n# Usage\niex> user = %User{name: \"John\", address: %{city: \"London\", street: %{name: \"High St\", house: 42}}}\niex> User.validate(user)\n{:ok, %User{...}}\n```","ref":"readme.html#nested-structures"},{"type":"extras","title":"Type System - Estructura","doc":"Estructura provides a rich type system with built-in types and scaffolds for custom types:\n\n#### Built-in Types\n\n- `DateTime` - For handling datetime values\n- `Date` - For date values\n- `Time` - For time values\n- `URI` - For URI handling\n- `IP` - For IPv4 and IPv6 addresses\n- `String` - For string values\n\n```elixir\ndefmodule Event do\n  use Estructura.Nested\n\n  defstruct [\n    timestamp: nil,\n    url: nil\n  ]\n\n  def type(:timestamp), do: Estructura.Nested.Type.DateTime\n  def type(:url), do: Estructura.Nested.Type.URI\nend\n```\n\n#### Type Scaffolds\n\n##### Enum Types\n\nCreate types with predefined values:\n\n```elixir\ndefmodule Status do\n  use Estructura.Nested.Type.Enum,\n    elements: [:pending, :active, :completed]\nend\n\niex> Status.validate(:pending)\n{:ok, :pending}\niex> Status.validate(:invalid)\n{:error, \"Expected :invalid to be one of: [:pending, :active, :completed]\"}\n```\n\n##### Tag Sets\n\nManage lists of predefined tags:\n\n```elixir\ndefmodule Categories do\n  use Estructura.Nested.Type.Tags,\n    elements: [:tech, :art, :science]\nend\n\niex> Categories.validate([:tech, :art])\n{:ok, [:tech, :art]}\niex> Categories.validate([:invalid])\n{:error, \"All tags are expected to be one of [:tech, :art, :science]...\"}\n```","ref":"readme.html#type-system"},{"type":"extras","title":"Coercion and Validation - Estructura","doc":"Estructura provides flexible coercion and validation:\n\n```elixir\ndefmodule Temperature do\n  use Estructura.Nested\n\n  defstruct value: 0, unit: :celsius\n\n  def coerce(:value, str) when is_binary(str) do\n    case Float.parse(str) do\n      {num, \"\"} -> {:ok, num}\n      _ -> {:error, \"Invalid number\"}\n    end\n  end\n\n  def validate(:value, v), do: v >= -273.15  # Absolute zero\n  def validate(:unit, u), do: u in [:celsius, :fahrenheit, :kelvin]\nend\n```","ref":"readme.html#coercion-and-validation"},{"type":"extras","title":"Lazy Values - Estructura","doc":"Use `Estructura.Lazy` for deferred computation:\n\n```elixir\ndefmodule Cache do\n  use Estructura.Nested\n\n  defstruct value: Estructura.Lazy.new(&expensive_computation/1)\n\n  def expensive_computation(_), do: :timer.sleep(1000) && :computed\nend\n```","ref":"readme.html#lazy-values"},{"type":"extras","title":"Flattening and Transformation - Estructura","doc":"Convert nested structures to flat representations:\n\n```elixir\ndefmodule User do\n  use Estructura.Nested, flattenable: true\n\n  defstruct name: \"\", address: %{city: \"\", postal_code: \"\"}\nend\n\niex> user = %User{name: \"John\", address: %{city: \"London\", postal_code: \"SW1\"}}\niex> Estructura.Flattenable.flatten(user)\n%{\"name\" => \"John\", \"address_city\" => \"London\", \"address_postal_code\" => \"SW1\"}\n```","ref":"readme.html#flattening-and-transformation"},{"type":"extras","title":"Property Testing - Estructura","doc":"Estructura supports property-based testing out of the box:\n\n```elixir\ndefmodule UserTest do\n  use ExUnit.Case\n  use ExUnitProperties\n\n  property \"valid users are validated\" do\n    check all %User{} = user <- User.__generator__() do\n      assert {:ok, ^user} = User.validate(user)\n    end\n  end\nend\n```","ref":"readme.html#property-testing"},{"type":"extras","title":"Changelog - Estructura","doc":"* `1.8.0` — `validate/1`\n* `1.7.0` — better infrastructure for `Types`, `URI`, `IP`, `Scaffold`\n* `1.6.0` — `jsonify: true | module()` option in a call to `Estructura.Flattenable.flatten/2`\n* `1.5.0` — no `:formulae` dependency\n* `1.4.1` — allow functions of arity 1 in `content` as coercers in a call to `Estructura.Aston.coerce/2`\n* `1.4.0` — allow coercers in a call to `Estructura.Aston.coerce/2`\n* `1.3.0` — calculated fields for `Estructura` and `Estructura.Nested`\n* `1.2.12` — export type from `Estructura.Nested`\n* `1.2.11` — nullable coercers\n* `1.2.10` — coercers for floats, and date/time values\n* `1.2.8` — `Estructura.Tree` → `Estructura.Aston` + `Aston.access/2` to retrieve and access key by names\n* `1.2.5` — `use Estructura.Nested flattenable: boolean(), jason: boolean(), transformer: boolean()`\n* `1.2.3` — Several `coerce/1` and `validate/1` clauses, default coercers\n* `1.2.2` — `Estructura.Flattenable`\n* `1.2.1` — Generators for `:datetime` and `:date`\n* `1.2.0` — `Estructura.Nested` would attempt to split keys by a delimiter if instructed\n* `1.1.0` — `Estructura.Aston` to hold an AST structure, like XML\n* `1.0.0` — Elixir v1.16 and deps\n* `0.6.0` — `Estructura.Transform` to produce squeezed representations of nested structs\n* `0.5.5` — export declarations of both `Estructura` and `Estructura.Nested` to docs\n* `0.5.4` — `Estructura.Nested` allows `cast/1` to cast nested structs from maps\n* `0.5.3` — `Estructura.diff/3` now understands maps\n* `0.5.2` — `Estructura.diff/3`\n* `0.5.1` — [BUG] Fixed `Collectable` and `Enumerable` injected implementations\n* `0.5.0` — `Estructura.Nested` for nested structures with validation, coercion, and generation\n* `0.4.2` — [BUG] Fixed wrong spec for `put!/3`\n* `0.4.1` — `Estructura.LazyMap.keys/1`, `Estructura.LazyMap.fetch_all/1`\n* `0.4.0` — `Estructura.Lazy`, `Estructura.LazyMap`\n* `0.3.2` — `put!/3`\n* `0.3.0` — `coercion` and `validation` are now injected as behaviours\n* `0.2.0` — `coercion`, `validation`, `put/3`\n\n## [Documentation](https://hexdocs.pm/estructura)","ref":"readme.html#changelog"},{"type":"extras","title":"Powerful Nested Structures","doc":"# Powerful Nested Structures","ref":"powerful-nested-structures.html"},{"type":"extras","title":"The Power of Structured Data - Powerful Nested Structures","doc":"In modern Elixir applications, handling complex nested data structures with proper validation, type coercion, and testing can be challenging. Enter Estructura, a powerful library that brings sophistication to Elixir structs while maintaining their elegant simplicity.","ref":"powerful-nested-structures.html#the-power-of-structured-data"},{"type":"extras","title":"Core Features - Powerful Nested Structures","doc":"","ref":"powerful-nested-structures.html#core-features"},{"type":"extras","title":"1. Sophisticated Nested Structures - Powerful Nested Structures","doc":"Estructura allows you to define and work with deeply nested structures while maintaining full type safety and validation:\n\n```elixir\ndefmodule User do\n  use Estructura.Nested\n\n  shape %{\n    name: :string,\n    address: %{\n      city: :string,\n      street: %{name: :string, house: Estructura.Nested.Type.Integer}\n    },\n    created_at: Estructura.Nested.Type.DateTime,\n    tags: Estructura.Nested.Type.Tags\n  ]\nend\n```","ref":"powerful-nested-structures.html#1-sophisticated-nested-structures"},{"type":"extras","title":"2. Smart Type System - Powerful Nested Structures","doc":"Built-in types with validation, coercion, and generation capabilities:\n\n- `Estructura.Nested.Type.DateTime` - For timestamp handling\n- `Estructura.Nested.Type.Date` - For date values\n- `Estructura.Nested.Type.Time` - For time values\n- `Estructura.Nested.Type.URI` - For URL/URI validation\n- `Estructura.Nested.Type.IP` - For IPv4/IPv6 addresses\n- `Estructura.Nested.Type.String` - For string handling\n- Custom type scaffolding with `Estructura.Nested.Type.{Tags, Enum}`\n\n```elixir\n# Define custom enum type\ndefmodule Status do\n  use Estructura.Nested.Type.Enum,\n    elements: [:pending, :active, :completed]\nend\n\n# Use in your structure\ndefmodule Task do\n  use Estructura.Nested\n  \n  shape %{title: :string, status: Status}\nend\n```","ref":"powerful-nested-structures.html#2-smart-type-system"},{"type":"extras","title":"3. Powerful Validation - Powerful Nested Structures","doc":"Validate entire structures with a single call:\n\n```elixir\n# Define validation rules\ndefmodule Order do\n  use Estructura.Nested\n\n  shape %{amount: :positive_integer, currency: :any}\n\n  validate do\n    def amount(value) when value > 0, do: {:ok, value}\n    def amount(value), do: {:error, value}\n\n    def currency(value) when value in ~w(USD EUR GBP), do: {:ok, value}\n    def currency(value), do: {:error, value}\n  end\nend\n\n# Validate instances\norder = %Order{amount: 100, currency: \"USD\"}\n{:ok, validated} = Order.validate(order)\n```","ref":"powerful-nested-structures.html#3-powerful-validation"},{"type":"extras","title":"4. Intelligent Coercion - Powerful Nested Structures","doc":"Automatically convert data to the right types:\n\n```elixir\ndefmodule Temperature do\n  use Estructura.Nested\n\n  defstruct [:value, :unit]\n\n  def coerce(:value, str) when is_binary(str) do\n    case Float.parse(str) do\n      {num, \"\"} -> {:ok, num}\n      _ -> {:error, \"Invalid number\"}\n    end\n  end\n\n  def coerce(:unit, unit) when is_binary(unit),\n    do: {:ok, String.to_existing_atom(unit)}\nend\n\n# Coercion happens automatically\n{:ok, temp} = Temperature.cast(%{\n  \"value\" => \"23.5\",\n  \"unit\" => \"celsius\"\n})\n```","ref":"powerful-nested-structures.html#4-intelligent-coercion"},{"type":"extras","title":"5. Property-Based Testing - Powerful Nested Structures","doc":"Generate valid test data automatically:\n\n```elixir\ndefmodule UserTest do\n  use ExUnit.Case\n  use ExUnitProperties\n\n  property \"valid users are validated\" do\n    check all %User{} = user <- User.__generator__() do\n      assert {:ok, ^user} = User.validate(user)\n    end\n  end\nend\n```","ref":"powerful-nested-structures.html#5-property-based-testing"},{"type":"extras","title":"6. Flexible Data Transformation - Powerful Nested Structures","doc":"Transform nested structures into different formats:\n\n```elixir\n# Flatten nested structures\nflattened = Estructura.Flattenable.flatten(user)\nassert flattened[\"address_city\"] == \"London\"\n\n# Transform to different representation\ntransformed = Estructura.Transformer.transform(user)\nassert transformed[:address][:city] == \"London\"\n```","ref":"powerful-nested-structures.html#6-flexible-data-transformation"},{"type":"extras","title":"Advanced Features - Powerful Nested Structures","doc":"","ref":"powerful-nested-structures.html#advanced-features"},{"type":"extras","title":"Lazy Evaluation - Powerful Nested Structures","doc":"Estructura supports lazy value computation with `Estructura.Lazy`:\n\n```elixir\ndefmodule Cache do\n  use Estructura.Nested\n\n  defstruct value: Estructura.Lazy.new(&expensive_computation/1)\n\n  def expensive_computation(_) do\n    # This computation only happens when the value is accessed\n    :timer.sleep(1000)\n    :computed_value\n  end\nend\n\n# Usage with LazyMap\ncache = LazyMap.new(\n  [\n    foo: Estructura.Lazy.new(&parse_int/1),\n    bar: Estructura.Lazy.new(&current_time/1, 100) # Expires after 100ms\n  ],\n  \"42\"\n)\n```","ref":"powerful-nested-structures.html#lazy-evaluation"},{"type":"extras","title":"Calculated Fields - Powerful Nested Structures","doc":"Support for computed fields based on other struct values:\n\n```elixir\ndefmodule Order do\n  use Estructura.Nested\n\n  defstruct [\n    items: [],\n    tax_rate: 0.20,\n    subtotal: 0.0,\n    calculated: %{\n      total: nil  # Calculated field\n    }\n  ]\n\n  def calculate(:total, %{subtotal: subtotal, tax_rate: rate}),\n    do: subtotal * (1 + rate)\nend\n```","ref":"powerful-nested-structures.html#calculated-fields"},{"type":"extras","title":"Aston Tree Structure - Powerful Nested Structures","doc":"Handle AST-like tree structures with the Aston feature:\n\n```elixir\ndefmodule XMLDoc do\n  use Estructura.Aston\n\n  defstruct content: []\nend\n\n# Build and transform tree structures\ndoc = %XMLDoc{\n  content: [\n    element: \"root\",\n    attributes: [id: \"main\"],\n    children: [\n      [element: \"child\", content: \"text\"]\n    ]\n  ]\n}\n```","ref":"powerful-nested-structures.html#aston-tree-structure"},{"type":"extras","title":"Auto-splitting Keys - Powerful Nested Structures","doc":"Automatically handle different key formats in input data:\n\n```elixir\ndefmodule UserPrefs do\n  use Estructura.Nested\n\n  defstruct [\n    display: %{\n      theme: :light,\n      font_size: 14\n    }\n  ]\nend\n\n# Both formats work automatically\n{:ok, prefs1} = UserPrefs.cast(%{\n  display: %{theme: :dark}\n})\n\n{:ok, prefs2} = UserPrefs.cast(%{\n  \"display_theme\" => \"dark\",\n  \"display_font_size\" => \"16\"\n}, split: true)\n```","ref":"powerful-nested-structures.html#auto-splitting-keys"},{"type":"extras","title":"Enhanced JSON Support - Powerful Nested Structures","doc":"Configure JSON encoding behavior per module:\n\n```elixir\ndefmodule Document do\n  use Estructura.Nested,\n    flattenable: true,   # Enable flattening\n    jason: true,         # Enable JSON encoding\n    transformer: true    # Enable transformation\n\n  defstruct [title: \"\", content: \"\"]\nend\n\n# Automatically handles JSON encoding/decoding\njson = Jason.encode!(%Document{title: \"Test\"})\n{:ok, doc} = Document.cast(Jason.decode!(json))\n```","ref":"powerful-nested-structures.html#enhanced-json-support"},{"type":"extras","title":"Latest Features (v1.7+) - Powerful Nested Structures","doc":"","ref":"powerful-nested-structures.html#latest-features-v1-7"},{"type":"extras","title":"1. Infrastructure Improvements - Powerful Nested Structures","doc":"The latest version (1.7.0) brings significant improvements to:\n- Types system architecture\n- URI handling capabilities\n- IP address support\n- Enhanced scaffolding system","ref":"powerful-nested-structures.html#1-infrastructure-improvements"},{"type":"extras","title":"2. Improved Type System - Powerful Nested Structures","doc":"The type system now includes:\n\n```elixir\n# IP Address handling\ndefmodule Server do\n  use Estructura.Nested\n\n  defstruct [:hostname, :ip]\n  \n  def type(:ip), do: Estructura.Nested.Type.IP\nend\n\n# Create with sigil\nserver = %Server{\n  hostname: \"web1\",\n  ip: ~IP[192.168.1.1]\n}\n\n# URI validation\ndefmodule Website do\n  use Estructura.Nested\n\n  defstruct [:url]\n  \n  def type(:url), do: Estructura.Nested.Type.URI\nend\n\n# Automatic validation and coercion\n{:ok, site} = Website.cast(%{\n  url: \"https://example.com\"\n})\n```","ref":"powerful-nested-structures.html#2-improved-type-system"},{"type":"extras","title":"3. Enhanced JSON Integration - Powerful Nested Structures","doc":"Version 1.6.0 adds flexible JSON transformation options:\n\n```elixir\n# Configure JSON transformation\nflattened = Estructura.Flattenable.flatten(\n  complex_struct,\n  jsonify: true  # Convert to JSON-friendly format\n)\n\n# Or use a custom module\nflattened = Estructura.Flattenable.flatten(\n  complex_struct,\n  jsonify: MyJSONFormatter\n)\n```","ref":"powerful-nested-structures.html#3-enhanced-json-integration"},{"type":"extras","title":"4. Improved Coercion System - Powerful Nested Structures","doc":"New coercion capabilities for various types:\n\n```elixir\ndefmodule Measurement do\n  use Estructura.Nested\n\n  defstruct [:value, :timestamp]\n\n  # Nullable coercer for timestamps\n  def type(:timestamp), do: Estructura.Coercers.NullableDatetime\n  \n  # Float coercion with validation\n  def type(:value), do: Estructura.Coercers.Float\nend\n\n# Both nil and valid dates work\n{:ok, m1} = Measurement.cast(%{\n  value: \"123.45\",\n  timestamp: nil\n})\n\n{:ok, m2} = Measurement.cast(%{\n  value: \"123.45\",\n  timestamp: \"2024-01-01T10:00:00Z\"\n})\n```","ref":"powerful-nested-structures.html#4-improved-coercion-system"},{"type":"extras","title":"5. Type Scaffolding Improvements - Powerful Nested Structures","doc":"Enhanced scaffolding system for custom types:\n\n```elixir\ndefmodule Priority do\n  use Estructura.Nested.Type.Enum,\n    elements: [:low, :medium, :high],\n    coercer: fn\n      str when is_binary(str) -> {:ok, String.to_existing_atom(str)}\n      atom when is_atom(atom) -> {:ok, atom}\n      other -> {:error, \"Invalid priority: #{inspect(other)}\"}\n    end\nend\n\ndefmodule Categories do\n  use Estructura.Nested.Type.Tags,\n    elements: [:feature, :bug, :docs],\n    coercer: fn\n      tags when is_list(tags) ->\n        {:ok, Enum.map(tags, &String.to_existing_atom/1)}\n      other ->\n        {:error, \"Invalid tags: #{inspect(other)}\"}\n    end\nend\n```","ref":"powerful-nested-structures.html#5-type-scaffolding-improvements"},{"type":"extras","title":"6. Calculated Fields - Powerful Nested Structures","doc":"Support for dynamically calculated fields:\n\n```elixir\ndefmodule Invoice do\n  use Estructura.Nested\n\n  defstruct [\n    items: [],\n    tax_rate: 0.21,\n    calculated: %{\n      subtotal: nil,\n      tax: nil,\n      total: nil\n    }\n  ]\n\n  def calculate(:subtotal, %{items: items}),\n    do: Enum.sum(Enum.map(items, & &1.price))\n    \n  def calculate(:tax, %{calculated: %{subtotal: sub}, tax_rate: rate}),\n    do: sub * rate\n    \n  def calculate(:total, %{calculated: %{subtotal: sub, tax: tax}}),\n    do: sub + tax\nend\n```\n\nThese improvements make Estructura an even more powerful tool for handling complex data structures in Elixir applications. The enhanced type system, improved coercion capabilities, and flexible JSON handling provide everything needed for sophisticated data management.","ref":"powerful-nested-structures.html#6-calculated-fields"},{"type":"extras","title":"Real-world Use Cases - Powerful Nested Structures","doc":"","ref":"powerful-nested-structures.html#real-world-use-cases"},{"type":"extras","title":"API Response Handling - Powerful Nested Structures","doc":"```elixir\ndefmodule APIResponse do\n  use Estructura.Nested\n\n  defstruct [\n    data: %{\n      items: [],\n      meta: %{\n        page: 1,\n        total: 0\n      }\n    },\n    status: :ok\n  ]\n\n  def type(:status), do: ResponseStatus  # Enum type\n  def type(\"data.meta.page\"), do: Integer\n  def type(\"data.meta.total\"), do: Integer\nend\n\n# Parse and validate API responses\n{:ok, response} = APIResponse.cast(json_data)\n```","ref":"powerful-nested-structures.html#api-response-handling"},{"type":"extras","title":"Complex Form Data - Powerful Nested Structures","doc":"```elixir\ndefmodule RegistrationForm do\n  use Estructura.Nested\n\n  defstruct [\n    user: %{\n      email: \"\",\n      password: \"\",\n      preferences: %{\n        notifications: true,\n        timezone: \"UTC\"\n      }\n    },\n    terms_accepted: false\n  ]\n\n  def validate(\"user.email\", email),\n    do: String.match?(email, ~r/^[^\\s@]+@[^\\s@]+\\.[^\\s@]+$/)\n    \n  def validate(\"user.password\", pass),\n    do: String.length(pass) >= 8\nend\n```","ref":"powerful-nested-structures.html#complex-form-data"},{"type":"extras","title":"Configuration Management - Powerful Nested Structures","doc":"```elixir\ndefmodule AppConfig do\n  use Estructura.Nested\n\n  defstruct [\n    database: %{\n      host: \"localhost\",\n      port: 5432,\n      pool_size: 10\n    },\n    cache: %{\n      ttl: 3600,\n      backend: :redis\n    }\n  ]\n\n  def type(\"database.port\"), do: Integer\n  def type(\"database.pool_size\"), do: Integer\n  def type(\"cache.ttl\"), do: Integer\n  def type(\"cache.backend\"), do: CacheBackend  # Enum type\nend\n```","ref":"powerful-nested-structures.html#configuration-management"},{"type":"extras","title":"Getting Started - Powerful Nested Structures","doc":"1. Add Estructura to your dependencies:\n\n```elixir\ndef deps do\n  [\n    {:estructura, \"~> 0.2\"},\n    …\n  ]\nend\n```\n\n2. Configuration (optional):\n\n```elixir\n# In config/config.exs\nconfig :estructura,\n  jsonify: true,         # Enable JSON-friendly transformations\n  transformer: true      # Enable struct transformations\n```\n\n3. Define your structures:\n\n```elixir\ndefmodule MyApp.User do\n  use Estructura.Nested\n\n  defstruct [\n    name: \"\",\n    email: \"\",\n    settings: %{\n      theme: :light,\n      language: \"en\"\n    }\n  ]\n\n  # Types\n  def type(\"settings.theme\"), do: Theme  # Enum type\n  \n  # Validation\n  def validate(:email, email),\n    do: String.match?(email, ~r/^[^\\s@]+@[^\\s@]+\\.[^\\s@]+$/)\n    \n  # Coercion\n  def coerce(:name, name) when is_binary(name),\n    do: {:ok, String.trim(name)}\nend\n```","ref":"powerful-nested-structures.html#getting-started"},{"type":"extras","title":"Why Choose Estructura? - Powerful Nested Structures","doc":"1. **Type Safety**: Built-in type system with validation and coercion\n2. **Nested Handling**: Elegant handling of deeply nested structures\n3. **Testing Support**: Automatic test data generation\n4. **Flexible Transformation**: Convert between different data representations\n5. **Production Ready**: Used in production applications","ref":"powerful-nested-structures.html#why-choose-estructura"},{"type":"extras","title":"Conclusion - Powerful Nested Structures","doc":"Estructura brings the power of sophisticated data handling to Elixir while maintaining the language's elegance and simplicity. Whether you're building APIs, handling form data, or managing complex configurations, Estructura provides the tools you need to work with structured data effectively.\n\nStart using Estructura today and experience the power of proper structure handling in your Elixir applications.","ref":"powerful-nested-structures.html#conclusion"},{"type":"extras","title":"Tests as First-Class Citizens","doc":"# Tests as First-Class Citizens\n\nFor more than twenty years, it has been standard practice in the industry to test code before deploying it to production. People have invented unit tests, acceptance tests, integration tests, property-based tests. People even came up with TDD to make sure the tests actually work. People invented mocks and contracts, finally (I highly recommend reading [this note by Valim](https://dev.to/plataformatec/mocks-and-explicit-contracts-5eap), it literally opened my eyes to what’s wrong with mocks in most cases).\n\nIn theory, everything looks great. Here’s a unit test:\n\n```elixir\ntest \"addition works for integers\" do\n  assert 3 + 4 == 7\n  refute 3 + 4 == 8\nend\n```\n\nIn practice, we usually test more complex data structures, and we are faced with the need to create those very data for testing. The code for each test turns into a huge chunk of boilerplate (imagine a JSON response from some external service, like a weather forecast). Lazy programmers invented data factories, like `Faker` and similar libraries. Smart programmers use the aforementioned property-based testing, but even there, you have to build a lot by hand.\n\nI’ve already complained that library authors rarely think about how to make it easier for users to test code that uses their API.","ref":"tests-as-first-class-citizens.html"},{"type":"extras","title":"A Lyrical Digression - Tests as First-Class Citizens","doc":"What words come to mind when we talk about data initialization? Validation, for sure. Here it’s appropriate to reference the brilliant text by Alexis King, [\"Parse Don’t Validate\"](https://lexi-lambda.github.io/blog/2019/11/05/parse-don-t-validate/). Coercion. Type coercion rules are needed; otherwise, any attempt to get a port address from an environment variable will be surrounded by ad-hoc conversions. In OOP, coercion and validation usually happen in setters. The functional approach leans more toward parametric polymorphism, coercing and validating in specially trained helpers. In any case, this field has been plowed three hundred times, and in every new project, adding such functionality can be done by an LLM with its eyes closed.\n\nBut to test the resulting object (data structure), you have to build a garden every time. It’s fine when `User` is just two fields: name and gender. But even adding a birth date brings the need to check edge cases: date in the future, age less than 18, and so on. In the real world, user data is 150 fields, often with multiple levels of nesting. Unit testing becomes almost meaningless, but even property-based tests require a huge amount of initialization code.\n\nSo at some point, I decided for myself that generating test data is as much an essential attribute of a data structure as validation and coercion out of the box. I hear an objection: production code should not carry anything extra just to simplify testing. It shouldn’t, and it doesn’t.\n\n`User.generate(options)` is a completely separate function that tests can call if needed. And REPL-driven development becomes much easier.\n\n## `Estructura`\n\nWhen I started working on the `estructura` library, a set of helpers for working with structures in Elixir, I knew that data generation would be a first-class citizen. To know what exactly to generate, you need at least types (at most-generation rules). But Elixir is a dynamically typed language. So I had to reinvent the wheel (spoiler: static types wouldn’t have helped me with generation, so the wheel turned out to be one of a kind). In the context of this library, a type means an implementation of the `Estructura.Nested.Type` behaviour. This is not an algebraic type or a set-theoretic type as in the Elixir core. These are utilitarian types, like a field type in a database.\n\nSo, to implement a type, we have the functions `coerce/1`, `validate/1`, and `generate/1`. For generating structure elements, parametric polymorphism is available, so to successfully generate structures like `%User{name: :string, %Address{street: :string}}`, it’s enough to have a generator for the `:string` type-everything else the library will do itself. Here’s what the `User` structure looks like in the library’s tests:\n\n```elixir\nuse Estructura.Nested\n\nshape %{\n  created_at: :datetime,\n  name: {:string, kind_of_codepoints: Enum.concat([?a..?c, ?l..?o])},\n  address: %{city: :string, street: %{name: [:string], house: :positive_integer}},\n  person: :string,\n  homepage: {:list_of, Estructura.Nested.Type.URI},\n  ip: Estructura.Nested.Type.IP,\n  data: %{age: :float},\n  birthday: Estructura.Nested.Type.Date,\n  title: {Estructura.Nested.Type.Enum, ~w|junior middle señor|},\n  tags: {Estructura.Nested.Type.Tags, ~w|backend frontend|}\n}\n```","ref":"tests-as-first-class-citizens.html#a-lyrical-digression"},{"type":"extras","title":"Coercion and Validation - Tests as First-Class Citizens","doc":"This is all well and good, but _Elixir_ is an immutable language. So you can’t just put all checks in a setter: there is no setter. It would seem this is a serious obstacle to an elegant solution, since requiring the user to call `Address.validate(address)` and then `User.validate(user)` is not an option. But here, one of the most underrated features in Elixir’s core comes to the rescue-lenses-`Access`, which allows you to update any hidden corners of a structure in one action through the core access functions `update_in/3` and friends. In fact, implementing `Access` for a structure is trivial, and as a reward, we get code where you can and should insert coercion and validation.\n\nAlso, through `Access`, the generated function `User.cast/2` works, which, for example, will distribute fields from a JSON received from a neighboring service.\n\nYes, setting values directly through `Map.put/3` or `%User{user | ...}` will break validation, but reflection in Java would bypass it too. Want to use a screwdriver? Plug it in and press the button, don’t bang the handle on the screw head.\n\nIt’s worth noting that, in addition to kosher coercion and validation via custom types, ad-hoc versions are also available (there’s no syntax error below; this is exactly how it works for nested fields: addressing via dot):\n\n```elixir\ncoerce do\n  def data.age(age) when is_float(age), do: {:ok, age}\n  def data.age(age) when is_integer(age), do: {:ok, 1.0 * age}\n  def data.age(age) when is_binary(age), do: {:ok, String.to_float(age)}\n  def data.age(age), do: {:error, \"Could not cast #{inspect(age)} to float\"}\nend\n```","ref":"tests-as-first-class-citizens.html#coercion-and-validation"},{"type":"extras","title":"Generation - Tests as First-Class Citizens","doc":"Generation is a bit more complicated. But recursion defeated the fear of unknown depths of nesting, and StreamData provided generators for all more or less usable types. I also added `URI`, `IP`, and meta-types `Enum` and `Tag`.\n\nNow, without any additional manipulation, we can define the `User` structure as shown above and try generating a couple of users.\n\n```elixir\niex|%_{}|1 ▶ Estructura.User.generator |> Stream.drop(10) |> Enum.take(1)\n\n%Estructura.User{\n  data: %Estructura.User.Data{age: 0.853515625},\n  name: \"cn\",\n  address: %Estructura.User.Address{\n    city: \"\",\n    street: %Estructura.User.Address.Street{\n      name: [\"ZkPjN\", \"tf\", \"J8\", \"kBA6iaQ\"],\n      house: 3\n    }\n  },\n  ip: ~IP[123.67.34.92],\n  title: \"junior\",\n  tags: [\"backend\", \"frontend\"],\n  person: \"cn, \",\n  created_at: ~U[2025-02-28 12:36:02Z],\n  homepage: [\n    %URI{\n      scheme: \"https\",\n      userinfo: nil,\n      host: \"example.com\",\n      port: 443,\n      path: \"/\",\n      query: \"bar=ok\",\n      fragment: \"anchor2\"\n    }\n  ],\n  birthday: ~D[2025-02-05]\n}\n```\n\nNow we can do property-based testing, for example, check that age must be greater than 18, or something like that.\n\nOf course, if there are additional requirements for field formats, the generator can be tweaked anywhere. You could even make it pull address data from `Faker`, but I personally highly discourage that; it’s better to see your code break on third-plane Unicode right away than to run into it in production.","ref":"tests-as-first-class-citizens.html#generation"},{"type":"extras","title":"Conclusion - Tests as First-Class Citizens","doc":"This note is not an advertisement for the library (though I do use it in every new project). I tried to show how we can significantly ease our users’ lives by providing testing tools out of the box. Doing this in 2025 is not just good manners, but a sign of developer maturity.","ref":"tests-as-first-class-citizens.html#conclusion"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}